import { Picker } from 'https://cdn.skypack.dev/emoji-picker-element';

/**
 * Reactions plugin for Converse.
 * https://xmpp.org/extensions/xep-0444.html
 */
converse.plugins.add('sib-reactions', {
  overrides: {
    ChatRoom: {
      async onMessage(attrs) {
        await this.onMessageWithReaction.apply(this, arguments);
        return await this.__super__.onMessage.apply(this, arguments);
      },
    },
    ChatBox: {
      async onMessage(attrs) {
        await this.onMessageWithReaction.apply(this, arguments);
        return await this.__super__.onMessage.apply(this, arguments);
      },
      async onMessageWithReaction(attrs) {
        attrs = await attrs;

        // Update message model with the reactions
        if (attrs.reaction_id) {
          const message = this.messages.findWhere({
            'id': attrs.reaction_id,
          });

          if (message) {
            const reactions = message.get('reactions') || {};

            // If there are reactions for this user, replace with the latest ones
            reactions[attrs.from_real_jid] = {
              emojis: attrs.reaction_emojis,
              origin_id: attrs.id,
              from_real_jid: attrs.from_real_jid,
            };

            message.set('reactions', {}); // Forces refresh
            message.save('reactions', reactions);
          }
        }
      },
    },
    ChatRoom: {
      async onMessage(attrs) {
        await this.onMessageWithReaction.apply(this, arguments);
        return await this.__super__.onMessage.apply(this, arguments);
      },
    },
  },
  initialize() {
    const _converse = this._converse;
    const { api, __ } = _converse;
    const { $msg, u, Strophe, sizzle, html, _ } = converse.env;

    Strophe.addNamespace('REACTIONS', 'urn:xmpp:reactions:0');

    /**
     * Send a reaction stanza for this user with the given emojis.
     * @param message
     * @param emojis
     * @param update
     */
    function sendReactionStanza(message, emojis, update = true) {
      const msgid = message.get('msgid');
      const type = message.get('type');
      const to = message.get(type === 'groupchat' ? 'from_muc' : 'contact_jid');

      const msg = $msg({
        'id': u.getUniqueId(),
        'to': to,
        'type': type,
      })
        .c('store', {
          'xmlns': Strophe.NS.HINTS,
        })
        .up()
        .c('reactions', {
          'id': msgid,
          'xmlns': Strophe.NS.REACTIONS,
        });

      // Insert reaction emojis into the stanza
      emojis.forEach(em => {
        msg.c('reaction').t(em).up();
      });

      // Send reaction stanza
      // TODO: Await?
      _converse.connection.send(msg);

      // Force UI update in the model
      if (update) {
        const from_real_jid = _converse.bare_jid;
        const reactions = message.get('reactions') || {};
        reactions[from_real_jid] = {
          emojis,
          origin_id: reactions[from_real_jid]?.origin_id || null,
          from_real_jid,
        };

        message.set('reactions', {}); // Forces refresh
        message.save('reactions', reactions);
      }

      // TODO: Remove
      console.log('sendReactionStanza', msg.toString());
    }

    /**
     * Get the current user emojis for a given message.
     * @param message
     */
    function getUserReactionEmojis(message) {
      const reactions = message.get('reactions') || {};
      const from_real_jid = _converse.bare_jid;
      const reaction = reactions[from_real_jid];
      return reaction?.emojis || [];
    }

    /**
     * Extends the Message element.
     * @type {*}
     */
    const MessageMixin = {
      renderMessageText() {
        return html`
          ${_renderMessageText.call(this)}
          ${this.hasReactions() ? this.renderReactions() : ''}
        `;
      },

      /**
       * Render the reactions list.
       * Adds an additional icon to add a reaction.
       * @returns {html}
       */
      renderReactions() {
        return html`
          <div class="chat-msg__reactions">
            ${this.getEmojisByCount().map(({ emoji, count, from_user }) => html`
              <a href="#" class="chat-msg__reaction ${from_user ? 'chat-msg__reaction--me' : ''}"
                 @click="${ev => this.toggleReaction(ev, emoji)}">
                <span class="emoji">${emoji}</span>
                <span class="count">${count}</span>
              </a>
            `)}
            <a href="#" @click="${this.openEmojiPicker}" class="chat-msg--react"><i class="fa fa-smile"></i></a>
          </div>
        `;
      },

      /**
       * Returns TRUE if the message has reactions.
       * Sometimes the reactions array contains empty reactions list
       * when a someone removes all of their emojis.
       * @returns {boolean}
       */
      hasReactions() {
        const reactions = this.model.get('reactions');
        return reactions ? Object.values(reactions).map(r => r.emojis).flat().length > 0 : false;
      },

      /**
       * Returns an object with the emoji as key and its count for value.
       * @returns {*}
       */
      getEmojisByCount() {
        const reactions = this.model.get('reactions');
        const from_real_jid = _converse.bare_jid;
        const emojis = {};

        for (let r of _.values(reactions)) {
          for (let e of r.emojis) {
            if (emojis.hasOwnProperty(e)) {
              emojis[e].count += 1;
              emojis[e].from_user = emojis[e].from_user || r.from_real_jid === from_real_jid;
            } else {
              emojis[e] = {
                emoji: e,
                count: 1,
                from_user: r.from_real_jid === from_real_jid,
              };
            }
          }
        }

        return _.values(emojis);
      },

      // TODO: Care about overflow according to the target position
      openEmojiPicker(ev) {
        ev.preventDefault();
        const self = this;
        const picker = new Picker();
        self.appendChild(picker);

        picker.addEventListener('emoji-click', function(e) {
          const emoji = e.detail.unicode;
          const emojis = getUserReactionEmojis(self.model);

          // TODO: If it has the reaction from this user already, remove it? (config)
          if (emojis?.includes(emoji)) {
            console.warn('The emoji is already in the list', emoji, emojis);
            picker.remove();
            return;
          }

          emojis.push(emoji);
          picker.remove();

          sendReactionStanza(self.model, emojis);
        });
      },

      /**
       * Add an emoji reaction on the message if another user added it.
       * Remove an emoji reaction from the message if the user added it.
       * @param ev
       * @param emoji
       */
      toggleReaction(ev, emoji) {
        ev.preventDefault();
        let emojis = getUserReactionEmojis(this.model);

        if (emojis?.includes(emoji)) {
          emojis = emojis.filter(em => em !== emoji);
        } else {
          emojis.push(emoji);
        }

        sendReactionStanza(this.model, emojis);
      },
    };

    // Apply message mixin
    const Message = _converse.api.elements.registry['converse-chat-message'];
    const _renderMessageText = Message.prototype.renderMessageText;
    Object.assign(Message.prototype, MessageMixin);

    /**
     * Register new reaction attributes.
     *
     * @param stanza
     * @param attrs
     * @returns {*}
     */
    function parseMessage(stanza, attrs) {
      const reactions = sizzle(`reactions[xmlns="${Strophe.NS.REACTIONS}"]`, stanza).pop();

      if (!!reactions) {
        attrs.reaction_id = reactions.getAttribute('id');
        attrs.reaction_emojis = Array.from(reactions.querySelectorAll('reaction')).map(el => el.textContent);
      }

      return attrs;
    }

    // Register handler for MUCs and chats
    api.listen.on('parseMessage', parseMessage);
    api.listen.on('parseMUCMessage', parseMessage);

    /**
     * Open the emoji drawer from message actions.
     * This is the first method of adding an emoji if there are none yet.
     */
    api.listen.on('getMessageActionButtons', (el, buttons) => {
      buttons.push({
        'i18n_text': __('React'),
        'handler': () => {
          const picker = new Picker();
          el.appendChild(picker);

          picker.addEventListener('emoji-click', function(e) {
            const emoji = e.detail.unicode;
            const emojis = getUserReactionEmojis(el.model);

            // TODO: If it has the reaction from this user already, remove it? (config)
            if (emojis?.includes(emoji)) {
              console.warn('The emoji is already in the list', emoji, emojis);
              picker.remove();
              return;
            }

            emojis.push(emoji);
            picker.remove();

            sendReactionStanza(el.model, emojis);
          });
        },
        'button_class': 'chat-msg__action-react',
        'icon_class': 'fa fa-smile',
        'name': 'react',
      });

      return buttons;
    });
  },
});
